local L = LibStub('AceLocale-3.0'):GetLocale('ProfessionsVault')
local addonName = L["ProfessionsVault"]
ProfessionsVault = LibStub("AceAddon-3.0"):NewAddon(addonName)
local addon = ProfessionsVault 
local AceGUI = LibStub("AceGUI-3.0")
local revision = tonumber(("$Revision: 16 $"):match("%d+"))
local DB_VERSION_MAJOR = 1
local DB_VERSION_MINOR = 1

local defaults = {
  profile = {
    debug = false, -- for addon debugging
    autoclose = false, -- close win after using a link
    tooltips = true, -- show tooltips in the ProfessionsVault window
    pos = {
      height = 432,
      width  = 510,
      left   = 350,
      top    = 823,
    },
    treestatus = {
      treewidth = 230,
    }
  }
}

local headerfont = GameFontHighlightLarge
local bodyfont = GameFontHighlightLarge
local profpages = false -- show a page for each profession
local activateontree = true


local primaryProf = { 
        2259, -- Alchemy
	2018, -- Blacksmithing
	7411, -- Enchanting
	4036, -- Engineering
	-- 9134, -- Herbalism
	45357, -- Inscription 
	25229, -- Jewelcrafting 
	2108, -- Leatherworking 
	-- 2575, -- Mining
	2656, -- Smelting
	3908, -- Tailoring
}
local secondaryProf = { 
	78670, -- Archaeology 
	2550, -- Cooking
	3273, -- First Aid
	-- 7620, -- Fishing
}
local function translateIDs(t) 
  local r = {}
  for _,v in pairs(t) do
    local name, _, icon = GetSpellInfo(v)
    if not name then
      debug("Failed to find spellID "..v)
    else
      r[name] = icon
    end
  end
  return r
end
primaryProf = translateIDs(primaryProf)
secondaryProf = translateIDs(secondaryProf)
local allProf = {}
local allProfSorted = {}
do 
  for p,v in pairs(primaryProf) do allProf[p]=v ; table.insert(allProfSorted, p) end
  table.sort(allProfSorted)
  local tmp = {}
  for p,v in pairs(secondaryProf) do allProf[p]=v ; table.insert(tmp, p)end
  table.sort(tmp)
  for _,v in ipairs(tmp) do table.insert(allProfSorted, v) end
  --myprint(allProfSorted)
end


local DB, DBc, settings
local charName

local function chatMsg(msg)
     DEFAULT_CHAT_FRAME:AddMessage(addonName..": "..msg)
end
local function debug(msg)
  if settings.debug then
     chatMsg(msg)
  end
end

local function myOptions() 
return {
  type = "group",
  args = {
    debug = {
      name = L["Debug"],
      desc = L["Toggle debugging output"],
      type = "toggle",
      set = function(info,val) settings.debug = val end,
      get = function(info) return settings.debug end
    },
    reset = {
      name = L["Reset"],
      desc = L["Reset the database"],
      type = "execute",
      func = function() addon:Reset(); end,
    },
    dump = {
      name = L["Dump"],
      desc = L["Dump the database to chat window"],
      type = "execute",
      func = function() addon:Dump()  end,
    },
    show = {
      name = L["Show"],
      desc = L["Show/Hide the ProfessionsVault window"],
      type = "execute",
      func = function() addon:ToggleWindow()  end,
    },
    autoclose = {
      name = L["Autoclose"],
      desc = L["Automatically close the ProfessionsVault window when opening a profession"],
      type = "toggle",
      set = function(info,val) settings.autoclose = val end,
      get = function(info) return settings.autoclose end
    },
    tooltips = {
      name = L["Tooltips"],
      desc = L["Show tooltips in the ProfessionsVault window"],
      type = "toggle",
      set = function(info,val) settings.tooltips = val end,
      get = function(info) return settings.tooltips end
    },
  }
} 
end

function addon:RefreshChar()
  DB.chars[charName] = DB.chars[charName] or {}
  DBc = DB.chars[charName]
  DBc.data = DBc.data or {}
  DBc.data.name = charName
  DBc.data.faction = UnitFactionGroup("player")
  DBc.data.level = UnitLevel("player")
  DBc.data.class = UnitClass("player")
  DBc.data.scanned = DBc.data.scanned or false
end

function addon:RefreshConfig()
  debug("RefreshConfig")
  charName = UnitName("player")
  DB = self.db.realm
  DB.chars = DB.chars or {}

  settings = addon.db.profile
  settings.vermaj = settings.vermaj or DB_VERSION_MAJOR
  settings.vermin = settings.vermin or DB_VERSION_MINOR

  addon:RefreshChar()
end

function addon:OnInitialize()
  self.db = LibStub("AceDB-3.0"):New("ProfessionsVaultDB", defaults)
  --LibStub("AceConfigRegistry-3.0"):ValidateOptionsTable(myOptions(), addonName)
  LibStub("AceConfig-3.0"):RegisterOptionsTable(addonName, myOptions(), {L["professionsvault"], L["pv"]})
  LibStub("AceEvent-3.0"):Embed(addon)
  settings = addon.db.profile

  debug("OnInitialize")

  self.db.RegisterCallback(self, "OnProfileChanged", "RefreshConfig")
  self.db.RegisterCallback(self, "OnProfileCopied", "RefreshConfig")
  self.db.RegisterCallback(self, "OnProfileReset", "RefreshConfig")
  self.db.RegisterCallback(self, "OnDatabaseReset", "RefreshConfig")
  self:RefreshConfig()
end

local LDB
function addon:OnEnable()
  debug("OnEnable")
  self:RegisterEvent("TRADE_SKILL_SHOW")

  --if not DBc.data.scanned then -- no professions registered on this toon
  --  addon:ScanProfessions()
  --end

  if LDB then
    return
  end
  LDB = LibStub:GetLibrary("LibDataBroker-1.1",true)
  if LDB then
    LDB:NewDataObject(addonName, {
        type = "launcher",
        label = addonName,
        icon = "Interface\\Icons\\spell_shadow_mindtwisting",
        OnClick = function(self, button)
                if button == "RightButton" then
                        --addon:OpenMenu(self,addon)
			do  end	
                else
                        addon:ToggleWindow()
                end
        end,
        OnTooltipShow = function(tooltip)
                if tooltip and tooltip.AddLine then
                        tooltip:SetText(addonName)
                        tooltip:AddLine(L["|cffff8040Left Click|r to toggle the ProfessionsVault window"])
                        tooltip:AddLine(L["|cffff8040Right Click|r for menu"])
                        tooltip:Show()
                end
        end,
     })
  end 
end

function addon:TRADE_SKILL_SHOW()
  debug("TRADE_SKILL_SHOW")
  addon:UpdateTrade()
end

function addon:UpdateTrade()
  local name, rank, rankmax = GetTradeSkillLine()
  local link = GetTradeSkillListLink()
  local linked, linkedName = IsTradeSkillLinked()
  if (linked) then
    debug("Ignoring "..linkedName.."'s "..name);
    return
  end
  if (name and name ~= "UNKNOWN" and link) then
    DBc[name] = {}

    DBc[name].link = link
    DBc[name].rank = rank
    DBc[name].rankmax = rankmax
    DBc[name].lastupdate = time()
    debug("Updated "..charName.."'s "..name);
  end
end

-- traversal function for character table
local cnext_sorted_names = {}
local function cnext(t,i)
   -- return them in reverse order
   if #cnext_sorted_names == 0 then 
     return nil
   else
      local n = cnext_sorted_names[#cnext_sorted_names]
      table.remove(cnext_sorted_names, #cnext_sorted_names)
      return n, t[n]
   end
end
local function cpairs(t)
  cnext_sorted_names = {}
  for n,_ in pairs(t) do
    table.insert(cnext_sorted_names, n)
  end
  table.sort(cnext_sorted_names, function (a,b) return b == charName or (a ~= charName and a > b) end)
  --myprint(cnext_sorted_names)
  return cnext, t, nil
end

function addon:Dump()
  for cname,dbc in cpairs(DB.chars) do
    local data = dbc.data or {}
    chatMsg("-=- "..cname.." "..(data.level or "").." "..(data.faction or "").." "..(data.class or "").." -=-")
    for _,pname in ipairs(allProfSorted) do
      if (dbc[pname]) then
        local pinfo = dbc[pname]
        local rank = pinfo.rank or "??"
        local rankmax = pinfo.rankmax or "??"
        chatMsg(" "..pinfo.link.."   ("..rank.."/"..rankmax..")")
      end
    end
  end
end

local function OnClose(widget)
  debug("OnClose")
  AceGUI:Release(widget)
  addon.gui = nil
end

local function cleanlink(link)
  local res = string.match(link,"|H(trade:[^|]+)|h")
  debug("cleanlink("..link..") => ".. res)
  return res
end

local function ActivateLink(cname,pname,link)
  debug("ActivateLink: "..cname.." "..pname)
  if IsShiftKeyDown() then
    local activeEditBox =  ChatEdit_GetActiveWindow();
    if activeEditBox then
      if (not ChatEdit_InsertLink(link) ) then
         debug("Failed to add tradeskill link to "..activeEditBox:GetName())
	 return
      end
    else
      return
    end
  else
    if cname == charName then
      CastSpellByName(pname)
    else
      local name, rank, rankmax = GetTradeSkillLine()
      local linked, linkedName = IsTradeSkillLinked()
      if linked and linkedName == cname and name and name == pname then -- already open so close it
        CloseTradeSkill()
      else -- open it (undocumented API - sshhh)
        SetItemRef(cleanlink(link),link,"LeftButton",ChatFrame1)
      end
    end
  end
  if settings.autoclose then
    addon.gui:Hide()
  end
end

function addon:ScanProfessions()
  debug("ScanProfessions")
  CloseTradeSkill()
  DB.chars[charName] = {}
  addon:RefreshChar()
  for p in pairs(allProf) do
    CastSpellByName(p) 
    CloseTradeSkill()
  end
  chatMsg(L["Profession Scan complete!"])
  DBc.data.scanned = true
  if addon.gui then
    addon:ToggleWindow()
    addon:ToggleWindow()
  end
end

local function ProfTooltip(frame, pname, pinfo)
   if not frame then
     debug("Missing frame on ProfTooltip")
     return
   end
   --GameTooltip:SetOwner(frame, "ANCHOR_CURSOR"); 
   GameTooltip:SetOwner(frame, "ANCHOR_BOTTOMRIGHT"); 
   GameTooltip:SetText(pname.." ("..pinfo.rank.." / "..pinfo.rankmax..")"); 
   if pinfo.lastupdate then
     GameTooltip:AddLine("|cff00ff00"..date(nil,pinfo.lastupdate).."|r")
   end
   GameTooltip:AddLine(L["|cffff8040Click|r to open"])
   GameTooltip:AddLine(L["|cffff8040Shift-Click|r to link in chat"])
   GameTooltip:Show() 
end

local function ProfIcon(cname,pname,pinfo)
  local icon = AceGUI:Create("Icon")
  icon:SetImage(allProf[pname])
  icon.frame:RegisterForClicks("AnyDown") -- cheat to get right-clicks too
  icon:SetCallback("OnClick", function(button) ActivateLink(cname,pname,pinfo.link) end)
  icon:SetCallback("OnEnter", function(button) ProfTooltip(button.frame, pname, pinfo) end)
  icon:SetCallback("OnLeave", function(button) GameTooltip:Hide() end)
  return icon
end

local function MakeButton(text, fn) 
  local button = AceGUI:Create("Button")
  button:SetText(text)
  button:SetWidth(200)
  button:SetCallback("OnClick",fn)
  return button
end

local function parsePath(path)
  local fullpath, _, cname, pname = string.find(path,"^([%w%s]+)\001([%w%s]+)$")
  if not fullpath then
    cname = path
  end
  return cname, pname
end

local function OnSelect(widget, event, value)
  local cname, pname = parsePath(value)
  if pname then
    debug("OnSelect: "..cname.." / "..pname)
    addon.gui:SetStatusText(cname.." / "..pname)
    settings.treepath = { cname , pname }
  else
    cname = value
    debug("OnSelect: "..cname)
    addon.gui:SetStatusText(cname)
    settings.treepath = { cname }
  end
  local dbc = DB.chars[cname]
  if not pname or not settings.profpages then
    widget:ReleaseChildren()
    widget:SetLayout("List") 
    -- body for a character
    local label = AceGUI:Create("Label")
    label:SetText(cname.."'s Professions\n")
    label:SetFullWidth(true)
    label:SetFontObject(headerfont)
    widget:AddChild(label)
    if cname == charName and (not dbc.data or not dbc.data.scanned) then
      widget:AddChild(MakeButton(L["Scan My Professions"], function () addon:ScanProfessions(); end))
      return
    end
    local scroller = AceGUI:Create("ScrollFrame")
    scroller:SetLayout("List")
    scroller:SetFullHeight(true)
    scroller:SetFullWidth(true)
    widget:AddChild(scroller)
    for i=1,2 do
      local label = AceGUI:Create("Label")
      label:SetText(i==1 and "Primary:\n" or "Secondary:\n")
      label:SetFontObject(bodyfont)
      scroller:AddChild(label)
      local body = AceGUI:Create("SimpleGroup")
      body:SetLayout("Flow")
      body:SetFullWidth(true)
      --body:SetAutoAdjustHeight(true)
      for pname in pairs((i==1 and primaryProf) or (i==2 and secondaryProf)) do
        if (dbc[pname]) then
	  local pinfo = dbc[pname]
          body:AddChild(ProfIcon(cname,pname,pinfo))
	end
      end
      scroller:AddChild(body)
    end
    if pname and activateontree then
      ActivateLink(cname,pname,DB.chars[cname][pname].link)
    end
    return
  end

 
  local pinfo = DB.chars[cname][pname]

  widget:ReleaseChildren()
  widget:SetLayout("List") 
  
  -- body for a particular profession
  local header = AceGUI:Create("SimpleGroup")
  header:SetFullWidth(true)
  header:SetLayout("Flow") 
  header:AddChild(ProfIcon(cname,pname,pinfo))
  local label = AceGUI:Create("Label")
  label:SetText(cname.."'s "..pname.."\n")
  label:SetFontObject(headerfont)
  header:AddChild(label)
  widget:AddChild(header)

  local body = AceGUI:Create("SimpleGroup")
  body:SetFullWidth(true)
  local skill = AceGUI:Create("Label")
  skill:SetText("Skill level: "..pinfo.rank.." / "..pinfo.rankmax.."\n")
  skill:SetFontObject(bodyfont)
  skill:SetFullWidth(true)
  body:AddChild(skill)
  local link = AceGUI:Create("InteractiveLabel")
  link:SetCallback("OnClick", function(button) ActivateLink(cname,pname,pinfo.link) end)
  link:SetText("Skill link: "..pinfo.link)
  link:SetFontObject(bodyfont)
  link:SetFullWidth(true)
  body:AddChild(link)
  widget:AddChild(body)
end

function addon:CreateWindow()
  if addon.gui then
    return
  end
  local f = AceGUI:Create("Frame")
  addon.gui = f
  f:SetTitle(L["ProfessionsVault"])
  f:SetStatusText(L["ProfessionsVault listing of your toon professions"])
  f:SetCallback("OnClose", OnClose)
  f:SetLayout("Flow")
  f.frame:SetClampedToScreen(true)
  settings.pos = settings.pos or {}
  f:SetStatusTable(settings.pos)
  f:SetAutoAdjustHeight(true)

  f:AddChild(MakeButton(L["Profession Pane"], function () ToggleSpellBook(BOOKTYPE_PROFESSION); end)) 
  f:AddChild(MakeButton(L["Scan My Professions"], function () addon:ScanProfessions(); end))

  local treedata = {}
  for cname,dbc in cpairs(DB.chars) do
    local ct = { value = cname, text = cname, 
                 icon = "Interface\\GossipFrame\\ActiveQuestIcon",
                 children = {} }
    --print(cname)
    if dbc.data and dbc.data.faction and dbc.data.faction == "Horde" then
      ct.icon = "Interface\\TargetingFrame\\UI-PVP-Horde"
    elseif dbc.data and dbc.data.faction and dbc.data.faction == "Alliance" then
      ct.icon = "Interface\\TargetingFrame\\UI-PVP-Alliance"
    end
    for _,pname in ipairs(allProfSorted) do
      if (dbc[pname]) then
        local pinfo = dbc[pname]
        table.insert(ct.children, { value = pname, icon = allProf[pname],
                                      text = pname.." ("..pinfo.rank.." / "..pinfo.rankmax..")" })
      end
    end
    table.insert(treedata, ct)
  end

  local tree = AceGUI:Create("TreeGroup")
  tree:SetFullWidth(true)
  tree:SetFullHeight(true)
  --tree:SetAutoAdjustHeight(true)
  tree:SetTree(treedata)
  tree:SetCallback("OnGroupSelected", OnSelect)
  tree:SetCallback("OnButtonEnter", function(button,event,path,frame) 
                   local cname, pname = parsePath(path)
		   if pname then
                     ProfTooltip(frame, pname, DB.chars[cname][pname]) 
		   elseif cname then
		     local data = DB.chars[cname].data
                     GameTooltip:SetOwner(frame, "ANCHOR_BOTTOMRIGHT");
                     GameTooltip:SetText(cname);
                     local text = ""
                     for _,stat in ipairs({"level","faction","class"}) do
                       if data and data[stat] then
                          text = text..data[stat].." "
                       end
                     end
                     if #text then
                        GameTooltip:AddLine(text)
                     end
                     GameTooltip:Show()
		   end
		   end)
  tree:SetCallback("OnButtonLeave", function() GameTooltip:Hide() end)
  tree:EnableButtonTooltips(false)
  f:AddChild(tree)
  if settings.treepath then
    local oldact = activateontree
    activateontree = false
    tree:SelectByPath(unpack(settings.treepath))
    activateontree = oldact
  end
  settings.treestatus = settings.treestatus or {}
  tree:SetStatusTable(settings.treestatus)
end

function addon:ToggleWindow()
  debug("ToggleWindow")

  if addon.gui then
    addon.gui:Hide()
  else
    addon:CreateWindow()
  end
end

function addon:Reset() 
  if addon.gui then
    addon:ToggleWindow()
  end
  StaticPopup_Show("PROFESSIONSVAULT_RESET")
end

local function ResetConfirmed()
  addon.db:ResetDB() -- triggers RefreshConfig
  chatMsg(L["Reset complete."]); 
end

StaticPopupDialogs["PROFESSIONSVAULT_RESET"] = {
  text = L["Are you sure you want to reset the ProfessionsVault database and wipe all data?"],
  button1 = OKAY,
  button2 = CANCEL,
  OnAccept = ResetConfirmed,
  timeout = 0,
  whileDead = true,
  hideOnEscape = true,
}

